<?php
if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

if ( !class_exists( 'WooCommerce_Customer_Manager_Functions' ) ) {

	class WooCommerce_Customer_Manager_Functions {
		public function __construct() {
			$this->settings = get_option( 'wpo_wccm_settings' );

			// Customer Search
			add_filter( 'get_search_query', array( $this, 'wc_customers_search_label' ) );
			add_filter( 'query_vars', array( $this, 'add_custom_query_var' ) );
			add_action( 'parse_query', array( $this, 'wc_customers_search_custom_fields' ) );

			// Filters
			add_filter( 'request', array( $this, 'request_query' ) );

			$this->register_wc_customers();
			$this->register_customer_status();

			// automatically add customer when order is placed
			add_action( 'woocommerce_checkout_order_processed', array( $this, 'get_customer_from_order' ) );

			// update when order saved or user data changed!
			add_action( 'save_post', array( $this,'update_from_order' ) );
			add_action( 'profile_update', array( $this,'update_from_profile' ), 10, 2 );
			add_action( 'woocommerce_customer_save_address', array( $this,'update_from_myaccount' ), 10, 2 );

			// Customer notes AJAX
			add_action( 'wp_ajax_wc_customers_add_customer_note', array( $this, 'add_customer_note_ajax' ) );
			add_action( 'wp_ajax_wc_customers_delete_customer_note', array( $this, 'delete_customer_note_ajax' ) );
			add_action( 'wp_ajax_wc_customers_save_customer_number', array( $this, 'save_customer_number_ajax' ) );

			// update customer card
			add_action( 'save_post', array( $this,'save_customer_card' ), 10, 2 );

			// update customer when order deleted
			add_action( 'before_delete_post', array( $this, 'post_deleted' ), 10, 1 );

			// run customer import when initiated by user
			if ( ! empty( $_GET['update_wc_customers'] ) ) {
				$imported_customers = $this->import_customers();
				$import_count = count($imported_customers);
				wp_redirect( add_query_arg( array( 'import_count' => $import_count, 'wc_customers_imported' => 'true' ), admin_url( 'edit.php?post_type=wc_customers' ) ) );
			}

			if ( ! empty( $_GET['ignore_wc_customers_import'] ) ) {
				update_option( 'wpo_wccm_ignore_import', true );
			}

			if (isset($_REQUEST['wccm_reset'])) {
				$this->reset();
			}

			if ( ! empty($this->settings['wcpdf_show'])) {
				add_action( 'wpo_wcpdf_after_order_data', array( $this, 'wcpdf_customer_number' ), 10, 2 );
			}

			// export to CSV
			add_action( 'wp_ajax_wpo_wcccm_export', array($this, 'export_customers') );

		}
		public function export_customers () {
			// Check the nonce
			if( empty( $_GET['action'] ) || ! is_user_logged_in() || !check_admin_referer( $_GET['action'] ) ) {
				wp_die( __( 'You do not have sufficient permissions to access this page.', 'wpo_wcpdf' ) );
			}

			$csv = $this->build_csv();

			$filename  = 'woocommerce-shipping-csv-export';
			$filename .= '-' . date('Y-m-d_H-i') . '.csv';

			header("Content-Description: File Transfer");
			header("Content-Disposition: attachment; filename=$filename");
			header("Content-Type: text/plain");
			echo $csv;
			die();
		}

		public function build_csv () {
			$csv_columns = array(
				'Customer #'				=> '',
				'Name'						=> '',
				'Billing Address'			=> '',
				'Shipping Address'			=> '',
			);

			$customers = $this->get_customers();

			$export = array();
			foreach ($customers as $customer) {
				$export[] = array(
					'Customer #'				=> $customer['number'],
					'Name'						=> $customer['name'],
					'Billing Address'			=> preg_replace('#<br\s*/?>#i', ", ", $customer['billing_address']),
					'Shipping Address'			=> preg_replace('#<br\s*/?>#i', ", ", $customer['shipping_address']),
				);
			}

			$delimiter = '","';

			// get headers
			$csv_headers = '"'.implode($delimiter, array_keys($csv_columns)).'"'."\n";

			// get data
			foreach ($export as $row) {
				$row = array_merge($csv_columns, $row);
				$rows[] = '"'.htmlspecialchars_decode(implode($delimiter, $row)).'"';
			}
			$csv_data = implode("\n", $rows);

			$csv = $csv_headers . $csv_data;
			return $csv;
		}

		/**
		 * Change the label when searching orders.
		 *
		 * @access public
		 * @param mixed $query
		 * @return string
		 */
		public function wc_customers_search_label( $query ) {
			global $pagenow, $typenow;

			if ( 'edit.php' != $pagenow ) {
				return $query;
			}

			if ( $typenow != 'wc_customers' ) {
				return $query;
			}

			if ( ! get_query_var( 'wc_customers_search' ) ) {
				return $query;
			}

			return wp_unslash( $_GET['s'] );
		}

		/**
		 * Query vars for custom searches.
		 *
		 * @access public
		 * @param mixed $public_query_vars
		 * @return array
		 */
		public function add_custom_query_var( $public_query_vars ) {
			$public_query_vars[] = 'wc_customers_search';

			return $public_query_vars;
		}

		/**
		 * Search custom fields as well as content.
		 *
		 * @access public
		 * @param WP_Query $wp
		 * @return void
		 */
		public function wc_customers_search_custom_fields( $wp ) {
			global $pagenow, $wpdb;

			if ( 'edit.php' != $pagenow || empty( $wp->query_vars['s'] ) || $wp->query_vars['post_type'] != 'wc_customers' ) {
				return;
			}

			$search_fields = array(
				'_customer_user',
				'_customer_username',
				'_customer_number',
				'_customer_number_raw',
				'_customer_orders',
				'_shipping_address',
				'_billing_address',
				'_billing_phone',
				'_email',
			);

			// Search orders
			$post_ids = array_unique( array_merge(
				$wpdb->get_col(
					$wpdb->prepare( "
						SELECT p1.post_id
						FROM {$wpdb->postmeta} p1
						INNER JOIN {$wpdb->postmeta} p2 ON p1.post_id = p2.post_id
						WHERE
							( p1.meta_key IN ('" . implode( "','", $search_fields ) . "') AND p1.meta_value LIKE '%%%s%%' )
						",
						esc_attr( $_GET['s'] ), esc_attr( $_GET['s'] ), esc_attr( $_GET['s'] )
					)
				)
			) );

			// Remove s - we don't want to search order name
			unset( $wp->query_vars['s'] );

			// so we know we're doing this
			$wp->query_vars['wc_customers_search'] = true;

			// Search by found posts
			$wp->query_vars['post__in'] = $post_ids;
		}

		/**
		 * Filters and sorting handler
		 * @param  array $vars
		 * @return array
		 */
		public function request_query( $vars ) {
			global $typenow, $wp_query;

			if ( 'wc_customers' === $typenow ) {
				// Sorting
				if ( isset( $vars['orderby'] ) ) {
					switch( $vars['orderby'] ) {
						case 'customer_number':
							$vars = array_merge( $vars, array(
								'meta_key' 	=> '_customer_number',
								'orderby' 	=> 'meta_value_num'
							) );
							break;
						case 'order_count':
							$vars = array_merge( $vars, array(
								'meta_key' 	=> '_order_count',
								'orderby' 	=> 'meta_value_num'
							) );
							break;
						case 'last_order':
							$vars = array_merge( $vars, array(
								'meta_key' 	=> '_last_order',
								'orderby' 	=> 'meta_value_num'
							) );
							break;
					}
				} else {
					// default sorting
					$vars = array_merge( $vars, array(
						'meta_key' 	=> '_customer_number',
						'orderby' 	=> 'meta_value_num'
					) );
				}

				// Status
				if ( ! isset( $vars['post_status'] ) ) {
					$vars['post_status'] = array( 'wccm-registered', 'wccm-guest' );
				}
			}

			return $vars;
		}
		public function register_wc_customers () {
			register_post_type( 'wc_customers',
				array(
					'labels' => array(
						'name' 					=> __( 'Customers', 'wpo_wccm' ),
						'singular_name' 		=> __( 'Customer', 'wpo_wccm' ),
						'add_new' 				=> __( 'Add New', 'wpo_wccm' ),
						'add_new_item' 			=> __( 'Add New Customer', 'wpo_wccm' ),
						'edit' 					=> __( 'Show details', 'wpo_wccm' ),
						'edit_item' 			=> __(  'Customer Details', 'wpo_wccm' ),
						'new_item' 				=> __( 'New Customer', 'wpo_wccm' ),
						'view' 					=> __( 'View', 'wpo_wccm' ),
						'view_item' 			=> __( 'View Customer', 'wpo_wccm' ),
						'search_items' 			=> __( 'Search Customers', 'wpo_wccm' ),
						'not_found' 			=> __( 'No Customers found', 'wpo_wccm' ),
						'not_found_in_trash' 	=> __( 'No Customers found in Trash', 'wpo_wccm' ),
						'parent' 				=> __( 'Parent Customer', 'wpo_wccm' ),
					),
					'capabilities' => array(
						'create_posts'    => 'do_not_allow', // Removes support for the "Add New" function
					),
					'map_meta_cap'        => true, // Set to false, if users are not allowed to edit/delete existing posts
					'description'         => __( 'This is where customers are stored.', 'wpo_wccm' ),
					'public'              => false,
					'show_ui'             => true,
					// 'capability_type'     => 'shop_order',
					'map_meta_cap'        => true,
					'publicly_queryable'  => false,
					'exclude_from_search' => true,
					'show_in_menu'        => current_user_can( 'manage_woocommerce' ) ? 'woocommerce' : true,
					'hierarchical'        => false,
					'show_in_nav_menus'   => false,
					'rewrite'             => false,
					'query_var'           => false,
					'supports'            => array( 'title', 'comments', 'custom-fields' ),
					'has_archive'         => false,
				)
			);
		}

		public function register_customer_status () {
			register_post_status( 'wccm-registered', array(
				'label'                     => __( 'Registered', 'wpo_wccm' ),
				'public'                    => true,
				'exclude_from_search'       => false,
				'show_in_admin_all_list'    => true,
				'show_in_admin_status_list' => true,
				'label_count'               => _n_noop( 'Registered <span class="count">(%s)</span>', 'Registered <span class="count">(%s)</span>', 'wpo_wccm' )
			) );
			register_post_status( 'wccm-guest', array(
				'label'                     => __( 'Guest', 'wpo_wccm' ),
				'public'                    => true,
				'exclude_from_search'       => false,
				'show_in_admin_all_list'    => true,
				'show_in_admin_status_list' => true,
				'label_count'               => _n_noop( 'Guest <span class="count">(%s)</span>', 'Guest <span class="count">(%s)</span>', 'wpo_wccm' )
			) );
		}

		public function import_customers () {
			// prevent function error (WC2.1.5 and older)
			if ( ! function_exists( 'woocommerce_reset_loop' ) ) {
				remove_filter( 'loop_end', 'woocommerce_reset_loop' );
			}

			@set_time_limit(0);

			$limit = 200;
			$offset = 0;
			$postlist = '';

			$imported_customers = array();
			while (TRUE) {
				// fix Post Types order bug
				remove_filter('posts_orderby', 'CPTOrderPosts', 99, 2);

				// Query posts
				$orders = new WP_Query( array(
					'posts_per_page'	=> $limit,
					'offset'			=> $offset,
					'post_type'			=> 'shop_order',
					'orderby'			=> 'ID',
					'order'				=> 'ASC', // process oldest first!
					'post_status'		=> array('any'),
					'fields' 			=> 'ids',
				));

				if (!$orders->have_posts()) break;
			
				foreach( $orders->posts as $order_id ):
					// add customer to wc_customers
					$customer_id = $this->get_customer_from_order( $order_id );

					if ( $customer_id === false ) {
						// no email set for order? either way we shouldn't do anything with the id.
						continue;
					}

					if (!isset($imported_customers[$customer_id])) {
						$imported_customers[$customer_id] = 1;
					} else {
						$imported_customers[$customer_id]++;
					}
				endforeach;
			
				$offset += $limit;
			} /*(TRUE)*/

			// restore Post Types order bug
			if ( function_exists('CPTOrderPosts') ) {
				add_filter('posts_orderby', 'CPTOrderPosts', 99, 2);
			}

			return $imported_customers;
		}

		public function get_customer_from_order( $order_id ) {
			// get order meta
			$order_meta = get_post_meta( $order_id );
			// flatten values
			foreach ($order_meta as $key => &$value) {
				$value = $value[0];
			}

			if ( isset($order_meta['_customer_id']) ) {
				// customer_id already stored, check if it still exists
				if ( get_post_type( $order_meta['_customer_id'] ) == 'wc_customers' && get_post_status( $order_meta['_customer_id'] ) != 'trash') {
					return $order_meta['_customer_id'];
				}
			}

			// bail if no email is set
			if ( !isset($order_meta['_billing_email']) || empty($order_meta['_billing_email']) ) {
				return false;
			}

			$customers = new WP_Query(array(
				'post_type'			=> 'wc_customers',
				'post_status'		=> array('any' ),
				'meta_query'		=> array(
					array(
						'key'		=> '_email',
						'compare'	=> 'LIKE',
						'value'		=> $order_meta['_billing_email'],
					),
				),
			));
			
			if (!$customers->have_posts()) {
				// customer not in Customer database yet!
				
				// get customer name
				$name = $this->get_customer_name ( $order_meta, 'order' );
				// add customer to wc_customers
				$new_customer = array(
					'post_title'	=> $name,
					'post_type'		=> 'wc_customers',
				);

				$customer_id = wp_insert_post( $new_customer );
				update_post_meta( $customer_id, '_email', $order_meta['_billing_email'] );

				// create customer number
				$settings = get_option( 'wpo_wccm_settings' );
				if ( !isset( $settings['next_number'] ) ) {
					$settings['next_number'] = 1001;
				}

				$customer_number_raw = apply_filters( 'wpo_wccm_create_customer_number', $settings['next_number'], $order_id, $customer_id );
				$customer_number = $this->get_customer_number( $customer_id, $customer_number_raw );
				// echo '<pre>';var_dump($customer_number_raw.'/'.$customer_number);echo '</pre>';die();
				update_post_meta( $customer_id, '_customer_number_raw', $customer_number_raw );
				update_post_meta( $customer_id, '_customer_number', $customer_number );

				// store number & id in user profile
				if ( isset( $order_meta['_customer_user'] ) && $order_meta['_customer_user'] != 0 ) {
					$user_id = $order_meta['_customer_user'];
					update_user_meta( $user_id, '_customer_id', $customer_id );
					update_user_meta( $user_id, '_customer_number_raw', $customer_number_raw );
					update_user_meta( $user_id, '_customer_number', $customer_number );
				}

				// increase next number
				$next_number = $customer_number_raw + 1;
				// make sure we're only increasing the next number if the new number is bigger!
				if ( $settings['next_number'] > $next_number ) {
					$next_number = $settings['next_number'];
				}
				$settings['next_number'] = $next_number;

				// save next number
				update_option( 'wpo_wccm_settings', $settings );
			} else {
				// get customer ID
				while($customers->have_posts()) {
					$customers->the_post();
					$customer_id = $customers->post->ID;
				}
				$customer_number_raw = $this->get_customer_number( $customer_id, '', false );
				$customer_number = $this->get_customer_number( $customer_id, $customer_number_raw );
			}

			// update customer meta
			$this->update_customer_meta ( $customer_id, $order_id, 'order' );

			// add customer_id & customer_number to order post_meta
			update_post_meta( $order_id, '_customer_id', $customer_id);
			update_post_meta( $order_id, '_customer_number_raw', $customer_number_raw );
			update_post_meta( $order_id, '_customer_number', $customer_number );

			// save last sync time
			update_option( 'wpo_wccm_last_sync', current_time('mysql') );

			return $customer_id;
		}

		public function get_customer_number ( $post_id, $customer_number_raw = '', $formatted = true ) {
			// determine what $id we're dealing with
			if( get_userdata( $post_id ) ) {
				$type = 'user';
			} else {
				$type = get_post_type( $post_id );
			}

			if (empty($customer_number_raw)) {
				// switch call based on type
				switch ($type) {
					case 'user':
						$customer_number_raw = get_user_meta( $post_id, '_customer_number_raw', true );
						if ( empty( $customer_number_raw ) ) {
							// backwards compatibility
							$customer_number_raw = get_user_meta( $post_id, '_customer_number', true );
						}
						break;
					case 'shop_order':
					case 'wc_customers':
					default:
						$customer_number_raw = get_post_meta( $post_id, '_customer_number_raw', true );
						if ( empty( $customer_number_raw ) ) {
							// backwards compatibility
							$customer_number_raw = get_post_meta( $post_id, '_customer_number', true );
						}
						break;
				}
			}
			$customer_number = $customer_number_raw;

			// return false if no number found
			if (empty($customer_number_raw)) {
				return false;
			}

			// return unformatted if required
			if ($formatted === false) {
				return $customer_number_raw;
			}

			// get prefix from settings
			$settings = get_option( 'wpo_wccm_settings' );
			if (!empty($settings['number_prefix'])) {
				$customer_number = $settings['number_prefix'].$customer_number;
			}

			return apply_filters( 'wpo_wccm_formatted_customer_number', $customer_number, $customer_number_raw, $post_id, $type );
		}

		public function get_customers () {
			$limit = 200;
			$offset = 0;
			$customers = array();

			while (TRUE) {
				// fix Post Types order bug
				remove_filter('posts_orderby', 'CPTOrderPosts', 99, 2);

				// Query posts
				$customer_posts = new WP_Query( array(
					'posts_per_page'	=> $limit,
					'offset'			=> $offset,
					'post_type'			=> 'wc_customers',
					'orderby'			=> 'ID',
					'order'				=> 'ASC', // oldest first!
					'post_status'		=> array('any'),
					'fields' 			=> 'ids',
				));

				if (!$customer_posts->have_posts()) break;
			
				foreach( $customer_posts->posts as $customer_id ) {
					// echo '<pre>';var_dump($customer_post);echo '</pre>';die();
					$customers[ $customer_id ] = array (
						'number'			=> get_post_meta( $customer_id, '_customer_number', true ),
						'name'				=> get_the_title( $customer_id ),
						'billing_address'	=> get_post_meta( $customer_id, '_billing_address', true ),
						'billing_phone'		=> get_post_meta( $customer_id, '_phone', true ),
						'billing_email'		=> get_post_meta( $customer_id, '_email', true ),
						'shipping_address'	=> get_post_meta( $customer_id, '_shipping_address', true ),
					);
				}

				$offset += $limit;
			} /*(TRUE)*/

			// restore Post Types order bug
			if ( function_exists('CPTOrderPosts') ) {
				add_filter('posts_orderby', 'CPTOrderPosts', 99, 2);
			}

			return $customers;
		}

		public function get_customer_orders ( $customer_email ) {
			// prevent function error (WC2.1.5 and older)
			if ( ! function_exists( 'woocommerce_reset_loop' ) ) {
				remove_filter( 'loop_end', 'woocommerce_reset_loop' );
			}

			$limit = 200;
			$offset = 0;
			$customer_orders = array();

			while (TRUE) {
				// fix Post Types order bug
				remove_filter('posts_orderby', 'CPTOrderPosts', 99, 2);

				// Query posts
				$orders = new WP_Query( array(
					'posts_per_page'	=> $limit,
					'offset'			=> $offset,
					'post_type'			=> 'shop_order',
					'orderby'			=> 'ID',
					'order'				=> 'ASC', // oldest first!
					'post_status'		=> array('any'),
					'fields' 			=> 'ids',
					'meta_key'			=> '_billing_email',
					'meta_value'		=> $customer_email,
				));

				if (!$orders->have_posts()) break;
			
				$customer_orders = array_merge($customer_orders, $orders->posts);
				$offset += $limit;
			} /*(TRUE)*/

			// restore Post Types order bug
			if ( function_exists('CPTOrderPosts') ) {
				add_filter('posts_orderby', 'CPTOrderPosts', 99, 2);
			}

			return $customer_orders;
		}

		public function update_customer_meta ( $customer_id, $object_id, $context ) {
			global $post_type;
			$settings = get_option( 'wpo_wccm_settings' );

			switch ($context) {
				case 'user':
					$user_id = $object_id;
					$user = get_user_by( 'id', $user_id );

					if ( empty( $customer_id ) ) {
						return;
					}

					// get user meta
					$user_meta = get_user_meta( $user_id );
					// flatten values
					foreach ($user_meta as $key => &$value) {
						$value = $value[0];
					}

					// get customer name
					$name = $this->get_customer_name ( $user_meta, 'user' );

					// update customer status (post_status) and title
					$status_update = array(
						'ID'			=> $customer_id,
						'post_title'	=> $name,
						'post_status'	=> 'wccm-registered',
					);

					// Fake wc_customers context to prevent issues with save_post hooks
					$last_post_type = $post_type;
					$post_type = 'wc_customers';
					wp_update_post( $status_update );
					$post_type = $last_post_type;

					update_post_meta( $customer_id, '_customer_username', $user->user_login );

					// phone number
					update_post_meta( $customer_id, '_billing_phone', $user_meta['_billing_phone'] );

					// billing address & shipping address
					$address_types = array( 'billing', 'shipping');
					foreach ($address_types as $type) {
						$address = apply_filters( 'woocommerce_my_account_my_address_formatted_address', array(
							'first_name'  => get_user_meta( $user_id, $type . '_first_name', true ),
							'last_name'   => get_user_meta( $user_id, $type . '_last_name', true ),
							'company'     => get_user_meta( $user_id, $type . '_company', true ),
							'address_1'   => get_user_meta( $user_id, $type . '_address_1', true ),
							'address_2'   => get_user_meta( $user_id, $type . '_address_2', true ),
							'city'        => get_user_meta( $user_id, $type . '_city', true ),
							'state'       => get_user_meta( $user_id, $type . '_state', true ),
							'postcode'    => get_user_meta( $user_id, $type . '_postcode', true ),
							'country'     => get_user_meta( $user_id, $type . '_country', true )
						), $user_id, $type );

						$formatted_address[$type] = WC()->countries->get_formatted_address( $address );
					}

					update_post_meta( $customer_id, '_billing_address', $formatted_address['billing'] );
					update_post_meta( $customer_id, '_shipping_address', $formatted_address['shipping'] );

					break;
				
				case 'order':
					$order_id = $object_id;
					$order = new WC_Order( $order_id );

					if ( empty( $customer_id ) ) {
						$customer_id = $this->get_customer_from_order( $order_id );
						if (!$customer_id) {
							return;
						}
					}

					// get order meta
					$order_meta = get_post_meta( $order_id );
					// flatten values
					foreach ($order_meta as $key => &$value) {
						$value = $value[0];
					}

					// update customer status (post_status) and title
					$status_update = array(
						'ID'			=> $customer_id,
						'post_status'	=> isset( $order_meta['_customer_user'] ) &&  $order_meta['_customer_user'] !=0 ? 'wccm-registered' : 'wccm-guest',
					);

					// name update if option set
					if ( isset( $settings['sync_orders'] ) ) {
						// get customer name
						$name = $this->get_customer_name ( $order_meta, 'order' );
						$status_update['post_title'] = $name;
					}

					// Fake wc_customers context to prevent issues with save_post hooks
					$last_post_type = $post_type;
					$post_type = 'wc_customers';
					wp_update_post( $status_update );
					$post_type = $last_post_type;

					// update list of customer orders
					if ( isset( $settings['connect_old_orders'] ) ) {
						// connect old orders
						$customer_email = get_post_meta( $order_id, '_billing_email', true );
						$customer_orders = $this->get_customer_orders ( $customer_email );
					} else {
						// just use what's stored
						$customer_orders = get_post_meta( $customer_id, '_customer_orders', true );
					}

					// prevent first empty result from saving
					if ( empty( $customer_orders ) ) {
						$customer_orders = array();
					} else {
						$customer_orders = (array) $customer_orders;
					}

					if ( !in_array( $order_id, $customer_orders ) && get_post_status( $order_id ) != 'trash' ) {
						$customer_orders[] = $order_id;						
					}

					// update last order if order_id is highest in order list
					if ( !empty( $customer_orders ) && $order_id >= max( $customer_orders ) && get_post_status( $order_id ) != 'trash' ) {
						update_post_meta( $customer_id, '_last_order', $order_id );
					}

					if ( empty( $customer_orders ) ) {
						update_post_meta( $customer_id, '_last_order', '' );
					}

					update_post_meta( $customer_id, '_customer_orders', $customer_orders );
					update_post_meta( $customer_id, '_order_count', count( $customer_orders ) );

					// user ID and username
					if ( isset( $order_meta['_customer_user'] ) && $order_meta['_customer_user'] != 0 ) {
						$user = get_user_by( 'id', $order_meta['_customer_user'] );
						update_post_meta( $customer_id, '_customer_user', $order_meta['_customer_user'] );
						update_post_meta( $customer_id, '_customer_username', $user->user_login );
					}

					if ( isset( $settings['sync_orders'] ) ) {
						// phone number
						$billing_phone = isset( $order_meta['_billing_phone'] ) ? $order_meta['_billing_phone'] : '';
						update_post_meta( $customer_id, '_billing_phone', $billing_phone );

						// billing address & shipping address
						update_post_meta( $customer_id, '_billing_address', $order->get_formatted_billing_address() );
						update_post_meta( $customer_id, '_shipping_address', $order->get_formatted_shipping_address() );
					}

					break;
				
			}
		}

		/**
		 * Update customer info when order is changed/saved
		 */
		public function update_from_order ( $order_id ) {
			global $post_type;
			if( $post_type == 'shop_order' ) {
				$customer_id = get_post_meta( $order_id, '_customer_id', true );
				$this->update_customer_meta ( $customer_id, $order_id, 'order' );
			}
			return $order_id;
		}

		/**
		 * Update customer info when user profile is changed/saved
		 */
		public function update_from_profile ( $user_id, $old_user_data ) {
			$customer_id = get_user_meta( $user_id, '_customer_id', true );
			$this->update_customer_meta ( $customer_id, $user_id, 'user' );
		}

		/**
		 * Update customer info when my account address is changed/saved
		 */
		public function update_from_myaccount ( $user_id, $load_address ) {
			$customer_id = get_user_meta( $user_id, '_customer_id', true );
			$this->update_customer_meta ( $customer_id, $user_id, 'user' );
		}

		/**
		 * Get full customer name from flattened user or post meta
		 * @param  array $post_meta EXPECTED TO BE FLATTENED!
		 */
		public function get_customer_name ( $post_meta, $context = 'order' ) {
			switch ($context) {
				case 'user':
					$prefix = 'billing_';
					break;
				case 'order':
					$prefix = '_billing_';
					break;
			}
			
			//build name safely 
			$first_name = isset($post_meta[$prefix.'first_name']) ? $post_meta[$prefix.'first_name'] : '';
			$last_name = isset($post_meta[$prefix.'last_name']) ? $post_meta[$prefix.'last_name'] : '';
			$email = isset($post_meta[$prefix.'email']) ? $post_meta[$prefix.'email'] : '';
			$name = trim($first_name.' '.$last_name); // remove leading & trailing whitespace

			// fallback to email if no name set (highly unlikely)
			if ( empty( $name ) ) {
				$name = $email;
			}
			
			return $name;
		}

		/**
		 * Update customer orders when order is deleted
		 */
		public function post_deleted ( $post_id ) {
			// first check if we're dealing with an order
			if ( get_post_type( $post_id ) == 'shop_order' ) {
				$customer_id = get_post_meta( $post_id, '_customer_id', true );
				$customer_orders = get_post_meta( $customer_id, '_customer_orders', true );
				$last_order = get_post_meta( $customer_id, '_last_order', true );

				// remove order from customer orders array
				if (in_array($post_id, $customer_orders)) {
					$customer_orders = array_diff( $customer_orders, array( $post_id ) );
					update_post_meta( $customer_id, '_customer_orders', $customer_orders );
				}

				// reset last order if it was the last order
				if ( $last_order == $post_id ) {
					if (!empty($customer_orders)) {
						update_post_meta( $customer_id, '_last_order', max( $customer_orders ) );
					} else {
						update_post_meta( $customer_id, '_last_order', '' );
					}
				}
			}
		}

		/**
		 * Adds a note (comment) to the customer
		 *
		 * @param string $note Note to add
		 * @return id Comment ID
		 */
		public function add_customer_note( $customer_id, $note ) {
			if ( is_user_logged_in() ) {
				$user                 = get_user_by( 'id', get_current_user_id() );
				$comment_author       = $user->display_name;
				$comment_author_email = $user->user_email;
			} else {
				$comment_author       = __( 'WC Customers', 'wpo_wccm' );
				$comment_author_email = get_bloginfo('admin_email');
				$comment_author_email = sanitize_email( $comment_author_email );
			}

			$comment_post_ID        = $customer_id;
			$comment_author_url     = '';
			$comment_content        = $note;
			$comment_agent          = 'WooCommerce';
			$comment_type           = 'customer_note';
			$comment_parent         = 0;
			$comment_approved       = 1;
			$commentdata            = compact( 'comment_post_ID', 'comment_author', 'comment_author_email', 'comment_author_url', 'comment_content', 'comment_agent', 'comment_type', 'comment_parent', 'comment_approved' );

			$comment_id = wp_insert_comment( $commentdata );

			return $comment_id;
		}

		/**
		 * Add customer note via ajax
		 */
		public static function add_customer_note_ajax() {
			global $wpo_wccm;

			check_ajax_referer( 'add-customer-note', 'security' );

			$post_id   = (int) $_POST['post_id'];
			$note      = wp_kses_post( trim( stripslashes( $_POST['note'] ) ) );

			if ( $post_id > 0 ) {
				$comment_id = $wpo_wccm->functions->add_customer_note( $post_id, $note );

				echo '<li rel="' . esc_attr( $comment_id ) . '" class="note">';
				echo '<div class="note_content">';
				echo wpautop( wptexturize( $note ) );
				echo '</div><p class="meta"><a href="#" class="delete_note">'.__( 'Delete note', 'woocommerce' ).'</a></p>';
				echo '</li>';
			}

			// Quit out
			die();
		}

		/**
		 * Delete order note via ajax
		 */
		public static function delete_customer_note_ajax() {
			check_ajax_referer( 'delete-customer-note', 'security' );

			$note_id = (int) $_POST['note_id'];

			if ( $note_id > 0 ) {
				wp_delete_comment( $note_id );
			}

			// Quit out
			die();
		}

		/**
		 * Save customer number via ajax
		 */
		public static function save_customer_number_ajax() {
			check_ajax_referer( 'save-customer-number', 'security' );

			$post_id = (int) $_POST['post_id'];
			$customer_number = wp_kses_post( trim( stripslashes( $_POST['customer_number'] ) ) );
			$original_number = wp_kses_post( trim( stripslashes( $_POST['original_number'] ) ) );

			// replace all occurrences of this customer number in database
			global $wpdb;

			// postmeta table
			$wpdb->query( 
				$wpdb->prepare( 
					"
					UPDATE $wpdb->postmeta
					SET meta_value = REPLACE (meta_value, '%s', '%s' )
					WHERE meta_key = '_customer_number_raw'
					AND meta_value = '%s'
					",
					$original_number, $customer_number, $original_number 
					)
			);

			// usermeta table
			$wpdb->query( 
				$wpdb->prepare( 
					"
					UPDATE $wpdb->usermeta
					SET meta_value = REPLACE (meta_value, '%s', '%s' )
					WHERE meta_key = '_customer_number_raw'
					AND meta_value = '%s'
					",
					$original_number, $customer_number, $original_number 
					)
			);

			// Quit out
			die();
		}


		public function save_customer_card ( $post_id, $post ) {
			global $post_type;
			// Check post type is wc_customer
			if ( $post_type != 'wc_customers' ) {
				return $post_id;
			}

			// If this is an autosave, our form has not been submitted, so we don't want to do anything.
			if ( defined( 'DOING_AUTOSAVE' ) && DOING_AUTOSAVE ) {
				return $post_id;
			}

			// Don't save revisions and autosaves
			if ( wp_is_post_revision( $post_id ) || wp_is_post_autosave( $post_id ) ) {
				return $post_id;
			}

			// get custom fields from settings
			$settings = get_option( 'wpo_wccm_settings' );
			if (!isset($settings['custom_fields']) || empty($settings['custom_fields'])) {
				return $post_id;
			}
			
			$custom_fieldnames = array();
			foreach ($settings['custom_fields'] as $custom_field ) {
				$custom_fieldnames[] = '_'.$custom_field['field'];
			}

			// allow custom metaboxes to hook in
			$custom_fieldnames = apply_filters( 'wpo_wccm_customer_card_fields', $custom_fieldnames );

			foreach ($custom_fieldnames as $type => $fieldname) {
				if ( isset( $_REQUEST['wpo_wccm'.$fieldname] ) ) {
					update_post_meta( $post_id, $fieldname, $_REQUEST['wpo_wccm'.$fieldname] );
				} else {
					delete_post_meta( $post_id, $fieldname );
				}
			}

			return $post_id;
		}

		/**
		 * WooCommerce PDF Invoices & Packing Slips compatibility
		 */
		public function wcpdf_customer_number ( $template_type, $order ) {
			global $wpo_wcpdf;
			$show_on = $this->settings['wcpdf_show'];
			if ( in_array( $template_type, array_keys( $show_on ) ) ) {
				$customer_number = $this->get_customer_number( $order->id );
				// try parent (original order for credit notes) if no number found
				if ( !$customer_number && $parent_order_id = wp_get_post_parent_id( $order->id ) ) {
					$customer_number = $this->get_customer_number( $parent_order_id );
				}
				?>
				<tr class="customer-number">
					<th><?php _e( 'Customer Number', 'wpo_wccm' ); ?>:</th>
					<td><?php echo $customer_number; ?></td>
				</tr>
				<?php
			}
		}

		public function reset(){
			global $wpdb;
			// remove customer ids and numbers from postmeta
			$wpdb->query(
				"
				DELETE FROM $wpdb->postmeta 
				WHERE meta_key in ('_customer_id', '_customer_number_raw', '_customer_number')
				"
			);

			// remove customer ids and numbers from postmeta
			$wpdb->query(
				"
				DELETE FROM $wpdb->usermeta 
				WHERE meta_key in ('_customer_id', '_customer_number_raw', '_customer_number')
				"
			);

			// remove all customers
			@set_time_limit(0);

			$limit = 200;
			$offset = 0;

			while (TRUE) {
				// fix Post Types order bug
				remove_filter('posts_orderby', 'CPTOrderPosts', 99, 2);

				// Query posts
				$customers = new WP_Query( array(
					'posts_per_page'	=> $limit,
					'offset'			=> $offset,
					'post_type'			=> 'wc_customers',
					'orderby'			=> 'ID',
					'order'				=> 'ASC', // process oldest first!
					'post_status'		=> array('any'),
					'fields' 			=> 'ids',
				));

				if (!$customers->have_posts()) break;
			
				foreach( $customers->posts as $customer_id ) {
				     wp_delete_post( $customer_id, true);
				}
			
				$offset += $limit;
			} /*(TRUE)*/

			// restore Post Types order bug
			if ( function_exists('CPTOrderPosts') ) {
				add_filter('posts_orderby', 'CPTOrderPosts', 99, 2);
			}
			
			// remove last sync date
			delete_option( 'wpo_wccm_last_sync' );
			delete_option( 'wpo_wccm_ignore_import' );

			// remove reset call when done!
			wp_redirect( remove_query_arg( 'wccm_reset' ) );
		}

	} // end class
} // end class_exists