<?php
/**
 * Plugin Name: WooCommerce Customer Manager
 * Plugin URI: http://www.wpovernight.com
 * Description: Simple customer manager for WooCommerce, for both guest & registered purchases
 * Version: 1.2.5
 * Author: Ewout Fernhout
 * Author URI: http://www.wpovernight.com
 * License: GPLv2 or later
 * License URI: http://www.opensource.org/licenses/gpl-license.php
 */

if ( !class_exists( 'WooCommerce_Customer_Manager' ) ) {

	class WooCommerce_Customer_Manager {
		public static $plugin_path;

		/**
		 * Constructor
		 */
		public function __construct() {
			self::$plugin_path = trailingslashit(dirname(__FILE__));

			// Init updater data
			$this->item_name	= 'WooCommerce Customer Manager';
			$this->file			= __FILE__;
			$this->license_slug	= 'wpo_wccm_license';
			$this->version		= '1.2.5';
			$this->author		= 'Ewout Fernhout';

			// load the localisation & classes
			add_action( 'plugins_loaded', array( $this, 'translations' ) ); // or use init?
			add_action( 'init', array( $this, 'load_classes' ) );

			// Load the updater
			add_action( 'init', array( $this, 'load_updater' ), 0 );

			// run lifecycle methods
			if ( is_admin() && ! defined( 'DOING_AJAX' ) ) {
				add_action( 'wp_loaded', array( $this, 'do_install' ) );
			}

		}

		/**
		 * Run the updater scripts from the WPO Sidekick
		 * @return void
		 */
		public function load_updater() {
			// Check if sidekick is loaded
			if (class_exists('WPO_Updater')) {
				$this->updater = new WPO_Updater( $this->item_name, $this->file, $this->license_slug, $this->version, $this->author );
			}
		}

		/**
		 * Load the translation / textdomain files
		 */
		public function translations() {
			load_plugin_textdomain( 'wpo_wccm', false, dirname( plugin_basename(__FILE__) ) . '/languages' );
		}

		/**
		 * Load the main plugin classes and functions
		 */
		public function includes() {
			include_once( 'includes/wccm-settings.php' );
			include_once( 'includes/wccm-writepanels.php' );
			include_once( 'includes/wccm-functions.php' );
		}
		

		/**
		 * Instantiate classes when woocommerce is activated
		 */
		public function load_classes() {
			if ( $this->is_woocommerce_activated() ) {
				$this->includes();
				$this->settings = new WooCommerce_Customer_Manager_Settings();
				$this->writepanels = new WooCommerce_Customer_Manager_Writepanels();
				$this->functions = new WooCommerce_Customer_Manager_Functions();
			} else {
				// display notice instead
				add_action( 'admin_notices', array ( $this, 'need_woocommerce' ) );
			}

		}

		/**
		 * Check if woocommerce is activated
		 */
		public function is_woocommerce_activated() {
			$blog_plugins = get_option( 'active_plugins', array() );
			$site_plugins = get_site_option( 'active_sitewide_plugins', array() );

			if ( in_array( 'woocommerce/woocommerce.php', $blog_plugins ) || isset( $site_plugins['woocommerce/woocommerce.php'] ) ) {
				return true;
			} else {
				return false;
			}
		}
		
		/**
		 * WooCommerce not active notice.
		 *
		 * @return string Fallack notice.
		 */
		 
		public function need_woocommerce() {
			$error = sprintf( __( 'WooCommerce Customer Manager requires %sWooCommerce%s to be installed & activated!' , 'wpo_wccm' ), '<a href="http://wordpress.org/extend/plugins/woocommerce/">', '</a>' );

			$message = '<div class="error"><p>' . $error . '</p></div>';
		
			echo $message;
		}

		/** Lifecycle methods *******************************************************
		 * Because register_activation_hook only runs when the plugin is manually
		 * activated by the user, we're checking the current version against the
		 * version stored in the database
		****************************************************************************/

		/**
		 * Handles version checking
		 */
		public function do_install() {
			$version_setting = 'wpo_wccm_version';
			$installed_version = get_option( $version_setting );

			// installed version lower than plugin version?
			if ( version_compare( $installed_version, $this->version, '<' ) ) {

				if ( ! $installed_version ) {
					$this->install();
				} else {
					$this->upgrade( $installed_version );
				}

				// new version number
				update_option( $version_setting, $this->version );
			}
		}


		/**
		 * Plugin install method. Perform any installation tasks here
		 */
		protected function install() {
			// stub
		}

		/**
		 * Plugin upgrade method.  Perform any required upgrades here
		 *
		 * @param string $installed_version the currently installed ('old') version
		 */
		protected function upgrade( $installed_version ) {
			$settings_key = 'wpo_wccm_settings';
			// 1.2.0 - set default for new settings
			if ( version_compare( $installed_version, '1.2.0', '<' ) ) {
				$current_settings = get_option( $settings_key );
				$new_defaults = array(
					'sync_orders'			=> 1,
					'connect_old_orders'	=> 1
				);
				
				$new_settings = array_merge($current_settings, $new_defaults);

				update_option( $settings_key, $new_settings );
			}
			// 1.2.5 - store raw customer number
			if ( version_compare( $installed_version, '1.2.5', '<' ) ) {
				@set_time_limit(0);
				// loop through customers
				$limit = 200;
				$offset = 0;
				while (TRUE) {
					// Query posts
					$customer_posts = new WP_Query( array(
						'posts_per_page'	=> $limit,
						'offset'			=> $offset,
						'post_type'			=> 'wc_customers',
						'orderby'			=> 'date',
						'order'				=> 'ASC', // oldest first!
						'post_status'		=> array('any'),
						'fields' 			=> 'ids',
					));

					if (!$customer_posts->have_posts()) break;
				
					foreach( $customer_posts->posts as $customer_id ) {
						$customer_number_raw = get_post_meta( $customer_id, '_customer_number_raw', true );
						if ( empty( $customer_number_raw ) ) {
							$customer_number = get_post_meta( $customer_id, '_customer_number', true );
							// remove any non-numerical characters just in case
							$customer_number_filtered = preg_replace('/[^0-9]/', '', $customer_number);
							// copy number to raw
							update_post_meta( $customer_id, '_customer_number_raw', $customer_number_filtered );
						}
					}

					$offset += $limit;
				} /*(TRUE)*/
			}			
		}

	} // class WooCommerce_Customer_Manager
} // class_exists

// Load main plugin class
$wpo_wccm = new WooCommerce_Customer_Manager();

/**
 * WPOvernight updater admin notice
 */
if ( ! class_exists( 'WPO_Updater' ) && ! function_exists( 'wpo_updater_notice' ) ) {

	if ( ! empty( $_GET['hide_wpo_updater_notice'] ) ) {
		update_option( 'wpo_updater_notice', 'hide' );
	}

	/**
	 * Display a notice if the "WP Overnight Sidekick" plugin hasn't been installed.
	 * @return void
	 */
	function wpo_updater_notice() {
		$wpo_updater_notice = get_option( 'wpo_updater_notice' );

		$blog_plugins = get_option( 'active_plugins', array() );
		$site_plugins = get_site_option( 'active_sitewide_plugins', array() );
		$plugin = 'wpovernight-sidekick/wpovernight-sidekick.php';

		if ( in_array( $plugin, $blog_plugins ) || isset( $site_plugins[$plugin] ) || $wpo_updater_notice == 'hide' ) {
			return;
		}

		echo '<div class="updated fade"><p>Install the <strong>WP Overnight Sidekick</strong> plugin to receive updates for your WP Overnight plugins - check your order confirmation email for more information. <a href="'.add_query_arg( 'hide_wpo_updater_notice', 'true' ).'">Hide this notice</a></p></div>' . "\n";
	}

	add_action( 'admin_notices', 'wpo_updater_notice' );
}
